// This file is part of Desktop App Toolkit,
// a set of libraries for developing nice desktop applications.
//
// For license and copyright information please follow this link:
// https://github.com/desktop-app/legal/blob/master/LEGAL
//

namespace Spelling::Helper {

[[nodiscard]] bool IsContextMenuTop(
	not_null<QMenu*> menu,
	QPoint mousePosition);

} // namespace Spelling::Helper
