// Microsoft Visual C++ generated resource script.
//

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#if defined(__MINGW64__) || defined(__MINGW32__)
	// MinGW-w64, MinGW
	#if defined(__has_include) && __has_include(<winres.h>)
		#include <winres.h>
	#else
		#include <afxres.h>
		#include <winresrc.h>
	#endif
#else
	// MSVC, Windows SDK
	#include <winres.h>
#endif

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON1               ICON                    "..\\art\\icon256.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 5,7,2,0
 PRODUCTVERSION 5,7,2,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x0L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "Telegram FZ-LLC"
            VALUE "FileDescription", "Telegram Desktop"
            VALUE "FileVersion", "5.7.2.0"
            VALUE "LegalCopyright", "Copyright (C) 2014-2024"
            VALUE "ProductName", "Telegram Desktop"
            VALUE "ProductVersion", "5.7.2.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////
