// This file is part of Desktop App Toolkit,
// a set of libraries for developing nice desktop applications.
//
// For license and copyright information please follow this link:
// https://github.com/desktop-app/legal/blob/master/LEGAL
//
#ifndef CONFIG_H
#define CONFIG_H

// enable threading
#define LOTTIE_THREAD_SUPPORT

//enable logging
//#define LOTTIE_LOGGING_SUPPORT

//enable module building of image loader
//#define LOTTIE_IMAGE_MODULE_SUPPORT

//enable lottie model caching
//#define LOTTIE_CACHE_SUPPORT

// disable image loader
#define LOTTIE_IMAGE_MODULE_DISABLED

#endif  // CONFIG_H
