/*
This file is part of Telegram Desktop,
the official desktop application for the Telegram messaging service.

For license and copyright information please follow this link:
https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
*/

#include <QtCore/QString>
#include <QtCore/QDateTime>

#include <rpl/rpl.h>

#include "base/basic_types.h"
