/*
This file is part of Telegram Desktop,
the official desktop application for the Telegram messaging service.

For license and copyright information please follow this link:
https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
*/
#include "ui/controls/window_outdated_bar.h"

namespace Ui {

object_ptr<RpWidget> CreateOutdatedBar(
		not_null<QWidget*> parent,
		const QString &workingPath) {
	return { nullptr };
}

} // namespace Ui
