#ifndef TGCALLS_FIELD_TRIALS_CONFIG_H
#define TGCALLS_FIELD_TRIALS_CONFIG_H

#include "api/transport/field_trial_based_config.h"

namespace tgcalls {

extern webrtc::FieldTrialBasedConfig fieldTrialsBasedConfig;

} // namespace tgcalls

#endif
